@@ -173,7 +173,8 @@ impl<'a> OciWriter<'a> {
     #[context("Writing OCI")]
     pub(crate) fn complete(&mut self) -> Result<()> {
         let utsname = nix::sys::utsname::uname();
-        let arch = MACHINE_TO_OCI[utsname.machine()];
+        let machine = utsname.machine();
+        let arch = MACHINE_TO_OCI.get(machine).unwrap_or(&machine);
 
         let rootfs_blob = self.root_layer.as_ref().unwrap();
         let root_layer_id = format!("sha256:{}", rootfs_blob.uncompressed_sha256);
@@ -310,7 +311,7 @@ impl<'a> std::io::Write for LayerWriter<'a> {
         self.compressor.write_all(srcbuf).unwrap();
         self.uncompressed_hash.update(srcbuf)?;
         let compressed_buf = self.compressor.get_mut().as_slice();
-        self.bw.write_all(&compressed_buf)?;
+        self.bw.write_all(compressed_buf)?;
         Ok(srcbuf.len())
     }
 
