@@ -3,10 +3,13 @@
 use crate::Result;
 use anyhow::{anyhow, Context};
 use camino::Utf8Path;
+use camino::Utf8PathBuf;
 use fn_error_context::context;
-use futures::prelude::*;
+use futures_util::TryFutureExt;
+use gio::glib;
 use gio::prelude::*;
-use glib::{Cast, Variant};
+use glib::Variant;
+use ostree::gio;
 use std::collections::HashMap;
 use std::convert::TryInto;
 use std::io::prelude::*;
@@ -23,21 +26,8 @@ const MAX_METADATA_SIZE: u32 = 10 * 1024 * 1024;
 /// https://stackoverflow.com/questions/258091/when-should-i-use-mmap-for-file-access
 const SMALL_REGFILE_SIZE: usize = 127 * 1024;
 
-// Variant formats, see ostree-core.h
-// TODO - expose these via introspection
-const OSTREE_COMMIT_FORMAT: &str = "(a{sv}aya(say)sstayay)";
-const OSTREE_DIRTREE_FORMAT: &str = "(a(say)a(sayay))";
-const OSTREE_DIRMETA_FORMAT: &str = "(uuua(ayay))";
-// const OSTREE_XATTRS_FORMAT: &str = "a(ayay)";
-
-/// State tracker for the importer.  The main goal is to reject multiple
-/// commit objects, as well as finding metadata/content before the commit.
-#[derive(Debug, PartialEq, Eq)]
-enum ImportState {
-    Initial,
-    Importing(String),
-}
-
+// The prefix for filenames that contain content we actually look at.
+const REPO_PREFIX: &str = "sysroot/ostree/repo/";
 /// Statistics from import.
 #[derive(Debug, Default)]
 struct ImportStats {
@@ -49,9 +39,9 @@ struct ImportStats {
 }
 
 /// Importer machine.
-struct Importer<'a> {
-    state: ImportState,
-    repo: &'a ostree::Repo,
+struct Importer {
+    repo: ostree::Repo,
+    remote: Option<String>,
     xattrs: HashMap<String, glib::Variant>,
     next_xattrs: Option<(String, String)>,
 
@@ -61,7 +51,7 @@ struct Importer<'a> {
     stats: ImportStats,
 }
 
-impl<'a> Drop for Importer<'a> {
+impl Drop for Importer {
     fn drop(&mut self) {
         let _ = self.repo.abort_transaction(gio::NONE_CANCELLABLE);
     }
@@ -91,20 +81,12 @@ fn header_attrs(header: &tar::Header) -> Result<(u32, u32, u32)> {
     Ok((uid, gid, mode))
 }
 
-fn format_for_objtype(t: ostree::ObjectType) -> Option<&'static str> {
-    match t {
-        ostree::ObjectType::DirTree => Some(OSTREE_DIRTREE_FORMAT),
-        ostree::ObjectType::DirMeta => Some(OSTREE_DIRMETA_FORMAT),
-        ostree::ObjectType::Commit => Some(OSTREE_COMMIT_FORMAT),
-        _ => None,
-    }
-}
-
 /// The C function ostree_object_type_from_string aborts on
 /// unknown strings, so we have a safe version here.
 fn objtype_from_string(t: &str) -> Option<ostree::ObjectType> {
     Some(match t {
         "commit" => ostree::ObjectType::Commit,
+        "commitmeta" => ostree::ObjectType::CommitMeta,
         "dirtree" => ostree::ObjectType::DirTree,
         "dirmeta" => ostree::ObjectType::DirMeta,
         "file" => ostree::ObjectType::File,
@@ -113,9 +95,8 @@ fn objtype_from_string(t: &str) -> Option<ostree::ObjectType> {
 }
 
 /// Given a tar entry, read it all into a GVariant
-fn entry_to_variant<R: std::io::Read>(
+fn entry_to_variant<R: std::io::Read, T: StaticVariantType>(
     mut entry: tar::Entry<R>,
-    vtype: &str,
     desc: &str,
 ) -> Result<glib::Variant> {
     let header = entry.header();
@@ -125,21 +106,82 @@ fn entry_to_variant<R: std::io::Read>(
     let n = std::io::copy(&mut entry, &mut buf)?;
     assert_eq!(n as usize, size);
     let v = glib::Bytes::from_owned(buf);
-    Ok(crate::variant_utils::variant_normal_from_bytes(vtype, v))
+    let v = Variant::from_bytes::<T>(&v);
+    Ok(v.normal_form())
 }
 
-impl<'a> Importer<'a> {
-    /// Import a commit object.  Must be in "initial" state.  This transitions into the "importing" state.
-    fn import_commit<R: std::io::Read>(
-        &mut self,
-        entry: tar::Entry<R>,
-        checksum: &str,
-    ) -> Result<()> {
-        assert_eq!(self.state, ImportState::Initial);
-        self.import_metadata(entry, checksum, ostree::ObjectType::Commit)?;
-        event!(Level::DEBUG, "Imported {}.commit", checksum);
-        self.state = ImportState::Importing(checksum.to_string());
-        Ok(())
+/// Parse an object path into (parent, rest, objtype).
+/// Normal ostree object paths look like 00/1234.commit.
+/// In the tar format, we may also see 00/1234.file.xattrs.
+fn parse_object_entry_path(path: &Utf8Path) -> Result<(&str, &Utf8Path, &str)> {
+    // The "sharded" commit directory.
+    let parentname = path
+        .parent()
+        .map(|p| p.file_name())
+        .flatten()
+        .ok_or_else(|| anyhow!("Invalid path (no parent) {}", path))?;
+    if parentname.len() != 2 {
+        return Err(anyhow!("Invalid checksum parent {}", parentname));
+    }
+    let name = path
+        .file_name()
+        .map(Utf8Path::new)
+        .ok_or_else(|| anyhow!("Invalid path (dir) {}", path))?;
+    let objtype = name
+        .extension()
+        .ok_or_else(|| anyhow!("Invalid objpath {}", path))?;
+    Ok((parentname, name, objtype))
+}
+
+fn parse_checksum(parent: &str, name: &Utf8Path) -> Result<String> {
+    let checksum_rest = name
+        .file_stem()
+        .ok_or_else(|| anyhow!("Invalid object path part {}", name))?;
+
+    if checksum_rest.len() != 62 {
+        return Err(anyhow!("Invalid checksum part {}", checksum_rest));
+    }
+    let checksum = format!("{}{}", parent, checksum_rest);
+    validate_sha256(&checksum)?;
+    Ok(checksum)
+}
+
+impl Importer {
+    fn new(repo: &ostree::Repo, remote: Option<String>) -> Self {
+        Self {
+            repo: repo.clone(),
+            remote,
+            buf: vec![0u8; 16384],
+            xattrs: Default::default(),
+            next_xattrs: None,
+            stats: Default::default(),
+        }
+    }
+
+    // Given a tar entry, filter it out if it doesn't start with the repository prefix.
+    // It is an error if the filename is invalid UTF-8.  If it is valid UTF-8, return
+    // an owned copy of the path.
+    fn filter_entry<R: std::io::Read>(
+        e: tar::Entry<R>,
+    ) -> Result<Option<(tar::Entry<R>, Utf8PathBuf)>> {
+        let orig_path = e.path()?;
+        let path = Utf8Path::from_path(&*orig_path)
+            .ok_or_else(|| anyhow!("Invalid non-utf8 path {:?}", orig_path))?;
+        // Ignore the regular non-object file hardlinks we inject
+        if let Ok(path) = path.strip_prefix(REPO_PREFIX) {
+            let path = path.into();
+            Ok(Some((e, path)))
+        } else {
+            Ok(None)
+        }
+    }
+
+    fn parse_metadata_entry(path: &Utf8Path) -> Result<(String, ostree::ObjectType)> {
+        let (parentname, name, objtype) = parse_object_entry_path(path)?;
+        let checksum = parse_checksum(parentname, name)?;
+        let objtype = objtype_from_string(objtype)
+            .ok_or_else(|| anyhow!("Invalid object type {}", objtype))?;
+        Ok((checksum, objtype))
     }
 
     /// Import a metadata object.
@@ -149,19 +191,23 @@ impl<'a> Importer<'a> {
         checksum: &str,
         objtype: ostree::ObjectType,
     ) -> Result<()> {
-        let vtype =
-            format_for_objtype(objtype).ok_or_else(|| anyhow!("Unhandled objtype {}", objtype))?;
-        let v = entry_to_variant(entry, vtype, checksum)?;
-        // FIXME insert expected dirtree/dirmeta
-        let _ = self
-            .repo
-            .write_metadata(objtype, Some(checksum), &v, gio::NONE_CANCELLABLE)?;
-        match objtype {
-            ostree::ObjectType::DirMeta => self.stats.dirmeta += 1,
-            ostree::ObjectType::DirTree => self.stats.dirtree += 1,
-            ostree::ObjectType::Commit => {}
-            _ => unreachable!(),
-        }
+        let v = match objtype {
+            ostree::ObjectType::DirTree => {
+                self.stats.dirtree += 1;
+                entry_to_variant::<_, ostree::TreeVariantType>(entry, checksum)?
+            }
+            ostree::ObjectType::DirMeta => {
+                self.stats.dirmeta += 1;
+                entry_to_variant::<_, ostree::DirmetaVariantType>(entry, checksum)?
+            }
+            o => return Err(anyhow!("Invalid metadata object type; {:?}", o)),
+        };
+        // FIXME validate here that this checksum was in the set we expected.
+        // https://github.com/ostreedev/ostree-rs-ext/issues/1
+        let actual =
+            self.repo
+                .write_metadata(objtype, Some(checksum), &v, gio::NONE_CANCELLABLE)?;
+        assert_eq!(actual.to_hex(), checksum);
         Ok(())
     }
 
@@ -294,21 +340,8 @@ impl<'a> Importer<'a> {
         entry: tar::Entry<'b, R>,
         path: &Utf8Path,
     ) -> Result<()> {
-        let parentname = path
-            .parent()
-            .map(|p| p.file_name())
-            .flatten()
-            .ok_or_else(|| anyhow!("Invalid path (no parent) {}", path))?;
-        if parentname.len() != 2 {
-            return Err(anyhow!("Invalid checksum parent {}", parentname));
-        }
-        let mut name = path
-            .file_name()
-            .map(Utf8Path::new)
-            .ok_or_else(|| anyhow!("Invalid path (dir) {}", path))?;
-        let mut objtype = name
-            .extension()
-            .ok_or_else(|| anyhow!("Invalid objpath {}", path))?;
+        let (parentname, mut name, mut objtype) = parse_object_entry_path(path)?;
+
         let is_xattrs = objtype == "xattrs";
         let xattrs = self.next_xattrs.take();
         if is_xattrs {
@@ -323,15 +356,7 @@ impl<'a> Importer<'a> {
                 .extension()
                 .ok_or_else(|| anyhow!("Invalid objpath {}", path))?;
         }
-        let checksum_rest = name
-            .file_stem()
-            .ok_or_else(|| anyhow!("Invalid objpath {}", path))?;
-
-        if checksum_rest.len() != 62 {
-            return Err(anyhow!("Invalid checksum rest {}", name));
-        }
-        let checksum = format!("{}{}", parentname, checksum_rest);
-        validate_sha256(&checksum)?;
+        let checksum = parse_checksum(parentname, name)?;
         let xattr_ref = if let Some((xattr_target, xattr_objref)) = xattrs {
             if xattr_target.as_str() != checksum.as_str() {
                 return Err(anyhow!(
@@ -348,39 +373,32 @@ impl<'a> Importer<'a> {
         } else {
             None
         };
-        let objtype = objtype_from_string(&objtype)
+        let objtype = objtype_from_string(objtype)
             .ok_or_else(|| anyhow!("Invalid object type {}", objtype))?;
-        match (objtype, is_xattrs, &self.state) {
-            (ostree::ObjectType::Commit, _, ImportState::Initial) => {
-                self.import_commit(entry, &checksum)
+        if is_xattrs && objtype != ostree::ObjectType::File {
+            return Err(anyhow!("Found xattrs for non-file object type {}", objtype));
+        }
+        match objtype {
+            ostree::ObjectType::Commit => {
+                return Err(anyhow!("Found multiple commit objects"));
             }
-            (ostree::ObjectType::File, true, ImportState::Importing(_)) => {
-                self.import_xattr_ref(entry, checksum)
-            }
-            (ostree::ObjectType::File, false, ImportState::Importing(_)) => {
-                self.import_content_object(entry, &checksum, xattr_ref)
-            }
-            (objtype, false, ImportState::Importing(_)) => {
-                self.import_metadata(entry, &checksum, objtype)
-            }
-            (o, _, ImportState::Initial) => {
-                return Err(anyhow!("Found content object {} before commit", o))
-            }
-            (ostree::ObjectType::Commit, _, ImportState::Importing(c)) => {
-                return Err(anyhow!("Found multiple commit objects; original: {}", c))
-            }
-            (objtype, true, _) => {
-                return Err(anyhow!("Found xattrs for non-file object type {}", objtype))
+            ostree::ObjectType::File => {
+                if is_xattrs {
+                    self.import_xattr_ref(entry, checksum)
+                } else {
+                    self.import_content_object(entry, &checksum, xattr_ref)
+                }
             }
+            objtype => self.import_metadata(entry, &checksum, objtype),
         }
     }
 
     /// Handle <checksum>.xattr hardlinks that contain extended attributes for
     /// a content object.
     #[context("Processing xattr ref")]
-    fn import_xattr_ref<'b, R: std::io::Read>(
+    fn import_xattr_ref<R: std::io::Read>(
         &mut self,
-        entry: tar::Entry<'b, R>,
+        entry: tar::Entry<R>,
         target: String,
     ) -> Result<()> {
         assert!(self.next_xattrs.is_none());
@@ -403,10 +421,6 @@ impl<'a> Importer<'a> {
 
     /// Process a special /xattrs/ entry (sha256 of xattr values).
     fn import_xattrs<R: std::io::Read>(&mut self, mut entry: tar::Entry<R>) -> Result<()> {
-        match &self.state {
-            ImportState::Initial => return Err(anyhow!("Found xattr object {} before commit")),
-            ImportState::Importing(_) => {}
-        }
         let checksum = {
             let path = entry.path()?;
             let name = path
@@ -439,13 +453,130 @@ impl<'a> Importer<'a> {
         Ok(())
     }
 
-    /// Consume this importer and return the imported OSTree commit checksum.
-    fn commit(mut self) -> Result<String> {
-        self.repo.commit_transaction(gio::NONE_CANCELLABLE)?;
-        match std::mem::replace(&mut self.state, ImportState::Initial) {
-            ImportState::Importing(c) => Ok(c),
-            ImportState::Initial => Err(anyhow!("Failed to find a commit object to import")),
+    fn import(mut self, archive: &mut tar::Archive<impl Read + Send + Unpin>) -> Result<String> {
+        self.repo.prepare_transaction(gio::NONE_CANCELLABLE)?;
+
+        // Create an iterator that skips over directories; we just care about the file names.
+        let mut ents = archive.entries()?.filter_map(|e| match e {
+            Ok(e) => {
+                if e.header().entry_type() == tar::EntryType::Directory {
+                    return None;
+                }
+                Self::filter_entry(e).transpose()
+            }
+            Err(e) => Some(Err(anyhow::Error::msg(e))),
+        });
+
+        // Read the commit object.
+        let (commit_ent, commit_path) = ents
+            .next()
+            .ok_or_else(|| anyhow!("Commit object not found"))??;
+
+        if commit_ent.header().entry_type() != tar::EntryType::Regular {
+            return Err(anyhow!(
+                "Expected regular file for commit object, not {:?}",
+                commit_ent.header().entry_type()
+            ));
         }
+        let (checksum, objtype) = Self::parse_metadata_entry(&commit_path)?;
+        if objtype != ostree::ObjectType::Commit {
+            return Err(anyhow!("Expected commit object, not {:?}", objtype));
+        }
+        let commit = entry_to_variant::<_, ostree::CommitVariantType>(commit_ent, &checksum)?;
+
+        let (next_ent, nextent_path) = ents
+            .next()
+            .ok_or_else(|| anyhow!("End of stream after commit object"))??;
+        let (next_checksum, next_objtype) = Self::parse_metadata_entry(&nextent_path)?;
+
+        if let Some(remote) = self.remote.as_deref() {
+            if next_checksum != checksum {
+                return Err(anyhow!(
+                    "Expected commitmeta checksum {}, found {}",
+                    checksum,
+                    next_checksum
+                ));
+            }
+            if next_objtype != ostree::ObjectType::CommitMeta {
+                return Err(anyhow!(
+                    "Using remote {} for verification; Expected commitmeta object, not {:?}",
+                    remote,
+                    objtype
+                ));
+            }
+            let commitmeta = entry_to_variant::<_, std::collections::HashMap<String, glib::Variant>>(
+                next_ent,
+                &next_checksum,
+            )?;
+
+            // Now that we have both the commit and detached metadata in memory, verify that
+            // the signatures in the detached metadata correctly sign the commit.
+            self.repo.signature_verify_commit_data(
+                remote,
+                &commit.data_as_bytes(),
+                &commitmeta.data_as_bytes(),
+                ostree::RepoVerifyFlags::empty(),
+            )?;
+
+            // Write the commit object, which also verifies its checksum.
+            let actual_checksum = self.repo.write_metadata(
+                objtype,
+                Some(&checksum),
+                &commit,
+                gio::NONE_CANCELLABLE,
+            )?;
+            assert_eq!(actual_checksum.to_hex(), checksum);
+            event!(Level::DEBUG, "Imported {}.commit", checksum);
+
+            // Finally, write the detached metadata.
+            self.repo.write_commit_detached_metadata(
+                &checksum,
+                Some(&commitmeta),
+                gio::NONE_CANCELLABLE,
+            )?;
+        } else {
+            // We're not doing any validation of the commit, so go ahead and write it.
+            let actual_checksum = self.repo.write_metadata(
+                objtype,
+                Some(&checksum),
+                &commit,
+                gio::NONE_CANCELLABLE,
+            )?;
+            assert_eq!(actual_checksum.to_hex(), checksum);
+            event!(Level::DEBUG, "Imported {}.commit", checksum);
+
+            // Write the next object, whether it's commit metadata or not.
+            let (meta_checksum, meta_objtype) = Self::parse_metadata_entry(&nextent_path)?;
+            match meta_objtype {
+                ostree::ObjectType::CommitMeta => {
+                    let commitmeta = entry_to_variant::<
+                        _,
+                        std::collections::HashMap<String, glib::Variant>,
+                    >(next_ent, &meta_checksum)?;
+                    self.repo.write_commit_detached_metadata(
+                        &checksum,
+                        Some(&commitmeta),
+                        gio::NONE_CANCELLABLE,
+                    )?;
+                }
+                _ => {
+                    self.import_object(next_ent, &nextent_path)?;
+                }
+            }
+        }
+
+        for entry in ents {
+            let (entry, path) = entry?;
+
+            if let Ok(p) = path.strip_prefix("objects/") {
+                self.import_object(entry, p)?;
+            } else if path.strip_prefix("xattrs/").is_ok() {
+                self.import_xattrs(entry)?;
+            }
+        }
+        self.repo.commit_transaction(gio::NONE_CANCELLABLE)?;
+
+        Ok(checksum)
     }
 }
 
@@ -459,51 +590,27 @@ fn validate_sha256(s: &str) -> Result<()> {
     Ok(())
 }
 
+/// Configuration for tar import.
+#[derive(Debug, Default)]
+pub struct TarImportOptions {
+    /// Name of the remote to use for signature verification.
+    pub remote: Option<String>,
+}
+
 /// Read the contents of a tarball and import the ostree commit inside.  The sha56 of the imported commit will be returned.
 #[instrument(skip(repo, src))]
 pub async fn import_tar(
     repo: &ostree::Repo,
     src: impl tokio::io::AsyncRead + Send + Unpin + 'static,
+    options: Option<TarImportOptions>,
 ) -> Result<String> {
+    let options = options.unwrap_or_default();
     let pipein = crate::async_util::async_read_to_sync(src);
     let repo = repo.clone();
     let import = tokio::task::spawn_blocking(move || {
-        let repo = &repo;
-        let mut importer = Importer {
-            state: ImportState::Initial,
-            repo,
-            buf: vec![0u8; 16384],
-            xattrs: Default::default(),
-            next_xattrs: None,
-            stats: Default::default(),
-        };
-        repo.prepare_transaction(gio::NONE_CANCELLABLE)?;
         let mut archive = tar::Archive::new(pipein);
-        for entry in archive.entries()? {
-            let entry = entry?;
-            if entry.header().entry_type() == tar::EntryType::Directory {
-                continue;
-            }
-            let path = entry.path()?;
-            let path = &*path;
-            let path = Utf8Path::from_path(path)
-                .ok_or_else(|| anyhow!("Invalid non-utf8 path {:?}", path))?;
-            let path = if let Ok(p) = path.strip_prefix("sysroot/ostree/repo/") {
-                p
-            } else {
-                continue;
-            };
-
-            if let Ok(p) = path.strip_prefix("objects/") {
-                // Need to clone here, otherwise we borrow from the moved entry
-                let p = &p.to_owned();
-                importer.import_object(entry, p)?;
-            } else if path.strip_prefix("xattrs/").is_ok() {
-                importer.import_xattrs(entry)?;
-            }
-        }
-
-        importer.commit()
+        let importer = Importer::new(&repo, options.remote);
+        importer.import(&mut archive)
     })
     .map_err(anyhow::Error::msg);
     let import: String = import.await??;
@@ -514,6 +621,19 @@ pub async fn import_tar(
 mod tests {
     use super::*;
 
+    #[test]
+    fn test_parse_metadata_entry() {
+        let c = "a8/6d80a3e9ff77c2e3144c787b7769b300f91ffd770221aac27bab854960b964";
+        let invalid = format!("{}.blah", c);
+        for &k in &["", "42", c, &invalid] {
+            assert!(Importer::parse_metadata_entry(k.into()).is_err())
+        }
+        let valid = format!("{}.commit", c);
+        let r = Importer::parse_metadata_entry(valid.as_str().into()).unwrap();
+        assert_eq!(r.0, c.replace('/', ""));
+        assert_eq!(r.1, ostree::ObjectType::Commit);
+    }
+
     #[test]
     fn test_validate_sha256() -> Result<()> {
         validate_sha256("a86d80a3e9ff77c2e3144c787b7769b300f91ffd770221aac27bab854960b964")?;
